#include "wind.h"
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>

#ifndef PAGE_SIZE
#define PAGE_SIZE 4096
#endif

int wind_connect(wind_client_t* client, const char* path, int blocking)
{
    memset(client, 0, sizeof *client);

    client->fd = socket(AF_UNIX, SOCK_STREAM, 0);
    if (client->fd < 0) return -1;

    struct sockaddr_un un;
    un.sun_family = AF_UNIX;
    strncpy(un.sun_path, path, sizeof(un.sun_path) - 1);

    if (connect(client->fd, (struct sockaddr*)&un, sizeof(un)) < 0)
    {
        close(client->fd);
        return -1;
    }

    if (!blocking) { fcntl(client->fd, F_SETFL, O_NONBLOCK); }

    fcntl(client->fd, F_SETFD, FD_CLOEXEC);

    return 0;
}

int wind_send_request(wind_client_t* client, uint8_t request_id, const void* request, size_t request_size)
{
    if (write(client->fd, &request_id, sizeof request_id) < 0) return -1;
    if (write(client->fd, request, request_size) < 0) return -1;
    return 0;
}

int wind_read_response(wind_client_t* client, uint8_t response_id, void* response, size_t response_size)
{
    int rc = read(client->fd, response, response_size);
    if (rc < 0)
    {
        if (errno == EAGAIN || errno == EINTR)
        {
            client->operation_in_progress = 1;
            client->saved_id = response_id;
            return 0;
        }
        else
            return -1;
    }
    return 1;
}

int wind_wait_for_sync_reply(wind_client_t* client, uint8_t response_id, void* response, size_t response_size)
{
    int max_other_messages = 5;

    while (max_other_messages)
    {
        uint8_t id;
        ssize_t rc = read(client->fd, &id, sizeof id);
        if (rc < 0 && (errno == EAGAIN || errno == EINTR)) continue;

        if (id == 0) // Error result
        {
            while (1)
            {
                int code;
                rc = read(client->fd, &code, sizeof code);
                if (rc < 0 && (errno == EAGAIN || errno == EINTR)) continue;
                errno = code;
                return -1;
            }
        }

        if (response_id != id)
        {
            client->handler(client, id, client->arg);
            max_other_messages--;
            continue;
        }

        while (1)
        {
            rc = read(client->fd, response, response_size);
            if (rc < 0 && (errno == EAGAIN || errno == EINTR)) continue;
            return 0;
        }
    }

    errno = ENOMSG;
    return -1;
}

int wind_client_fd(wind_client_t* client)
{
    return client->fd;
}

int wind_disconnect(wind_client_t* client)
{
    int rc = close(client->fd);
    memset(client, 0, sizeof *client);
    return rc;
}

void wind_set_handler(wind_client_t* client, int (*handler)(wind_client_t*, uint8_t, void*), void* arg)
{
    client->handler = handler;
    client->arg = arg;
}

int wind_check_for_messages(wind_client_t* client)
{
    if (client->operation_in_progress)
    {
        client->operation_in_progress = 0;
        client->handler(client, client->saved_id, client->arg);
        return 0;
    }

    uint8_t id;
    ssize_t rc = read(client->fd, &id, sizeof id);

    if (rc < 0)
    {
        if (errno == EAGAIN) return 0; // No messages, and the caller does not want us to block.
        if (errno == EINTR) return 0; // Let the caller check for anything having happened because a signal handler ran.
        return -1;
    }

    client->handler(client, id, client->arg);
    return 0;
}

static int adopt_shared_memory(const char* path, size_t size, void** ptr)
{
    int fd = shm_open(path, O_RDWR, 0660);
    if (fd < 0) return -1;

    void* p = mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
    if (p == MAP_FAILED)
    {
        int olderr = errno;
        close(fd);
        errno = olderr;
        return -1;
    }

    close(fd);
    *ptr = p;

    return 0;
}

int wind_create_window(wind_client_t* client, wind_window_t* window, struct wind_rect rect)
{
    memset(window, 0, sizeof *window);

    struct wind_create_window_request request;
    memcpy(&request.rect, &rect, sizeof rect);

    wind_send_request(client, wind_CreateWindow, &request, sizeof request);

    struct wind_create_window_response response;

    int result = wind_wait_for_sync_reply(client, wind_CreateWindowResponse, &response, sizeof response);
    if (result < 0) return -1;

    window->id = response.window;

    char path[WIND_STRING_LENGTH + 1];
    memset(path, 0, sizeof path);
    strncpy(path, response.shm_path, WIND_STRING_LENGTH);

    result = adopt_shared_memory(path, rect.height * rect.width * sizeof(int), &window->canvas);
    if (result < 0)
    {
        wind_close_window(client, window);
        return -1;
    }

    memcpy(&window->rect, &rect, sizeof rect);

    struct wind_remove_shared_request shm_request = {
        .window = response.window,
    };

    wind_send_request(client, wind_RemoveShared, &shm_request, sizeof shm_request);

    return 0;
}

int wind_close_window(wind_client_t* client, wind_window_t* window)
{
    struct wind_close_window_request request = {
        .window = window->id,
    };

    int result = wind_send_request(client, wind_CloseWindow, &request, sizeof request);

    if (window->canvas) munmap(window->canvas, window->rect.width * window->rect.height * sizeof(int));

    return result;
}

int wind_invalidate(wind_client_t* client, wind_window_t* window)
{
    struct wind_invalidate_request request = {
        .window = window->id,
    };

    return wind_send_request(client, wind_Invalidate, &request, sizeof request);
}

int wind_set_window_title(wind_client_t* client, wind_window_t* window, const char* title)
{
    struct wind_set_window_title_request request = {
        .window = window->id,
    };

    strncpy(request.title, title, WIND_STRING_LENGTH - 1);

    return wind_send_request(client, wind_SetWindowTitle, &request, sizeof request);
}

int wind_set_titlebar_height(wind_client_t* client, wind_window_t* window, int height)
{
    struct wind_set_titlebar_height_request request = {
        .window = window->id,
        .height = height,
    };

    return wind_send_request(client, wind_SetTitlebarHeight, &request, sizeof request);
}

int wind_get_screen_dimensions(wind_client_t* client, struct wind_rect* rect)
{
    struct wind_get_screen_rect_request request;
    struct wind_get_screen_rect_response response;

    wind_send_request(client, wind_GetScreenRect, &request, sizeof request);

    int result = wind_wait_for_sync_reply(client, wind_GetScreenRectResponse, &response, sizeof response);

    memcpy(rect, &response.rect, sizeof *rect);

    return result;
}

int wind_set_special_window_attributes(wind_client_t* client, wind_window_t* window,
                                       enum wind_window_attributes attributes)
{
    struct wind_set_special_window_attributes_request request = {
        .window = window->id,
        .attributes = attributes,
    };

    return wind_send_request(client, wind_SetSpecialWindowAttributes, &request, sizeof request);
}